@@ -3,12 +3,12 @@
  * pgtclCmds.c
  *	  C functions which implement pg_* tcl commands
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpgtcl/pgtclCmds.c,v 1.60 2001/12/03 14:49:46 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/interfaces/libpgtcl/Attic/pgtclCmds.c,v 1.69 2002/10/17 14:51:50 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -399,7 +399,6 @@ Pg_connect(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 int
 Pg_disconnect(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 {
-	Pg_ConnectionId *connid;
 	PGconn	   *conn;
 	Tcl_Channel conn_chan;
 
@@ -413,15 +412,14 @@ Pg_disconnect(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 	if (conn_chan == NULL)
 	{
 		Tcl_ResetResult(interp);
-		Tcl_AppendResult(interp, argv[1], " is not a valid connection\n", 0);
+		Tcl_AppendResult(interp, argv[1], " is not a valid connection", 0);
 		return TCL_ERROR;
 	}
 
-#if TCL_MAJOR_VERSION >= 8
-	conn = PgGetConnectionId(interp, argv[1], &connid);
-	if (connid->notifier_channel != NULL)
-		Tcl_UnregisterChannel(interp, connid->notifier_channel);
-#endif
+	/* Check that it is a PG connection and not something else */
+	conn = PgGetConnectionId(interp, argv[1], (Pg_ConnectionId **) NULL);
+	if (conn == (PGconn *) NULL)
+		return TCL_ERROR;
 
 	return Tcl_UnregisterChannel(interp, conn_chan);
 }
@@ -557,7 +555,8 @@ Pg_result(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 	result = PgGetResultId(interp, argv[1]);
 	if (result == (PGresult *) NULL)
 	{
-		Tcl_AppendResult(interp, argv[1], " is not a valid query result", 0);
+		Tcl_AppendResult(interp, "\n",
+						 argv[1], " is not a valid query result", 0);
 		return TCL_ERROR;
 	}
 
@@ -909,7 +908,7 @@ Pg_execute(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 
 		sprintf(oid_buf, "%u", PQoidValue(result));
 		if (Tcl_SetVar(interp, oid_varname, oid_buf,
-					   TCL_LEAVE_ERR_MSG) != TCL_OK)
+					   TCL_LEAVE_ERR_MSG) == NULL)
 		{
 			PQclear(result);
 			return TCL_ERROR;
@@ -1579,7 +1578,11 @@ Pg_lo_import(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 	lobjId = lo_import(conn, filename);
 	if (lobjId == InvalidOid)
 	{
-		sprintf(interp->result, "Pg_lo_import of '%s' failed", filename);
+		/*
+		 * What is the maximum size of this? FIXME if this is not a good
+		 * quess
+		 */
+		snprintf(interp->result, 128, "Pg_lo_import of '%s' failed", filename);
 		return TCL_ERROR;
 	}
 	sprintf(interp->result, "%u", lobjId);
@@ -1875,6 +1878,7 @@ Pg_listen(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 		notifies = (Pg_TclNotifies *) ckalloc(sizeof(Pg_TclNotifies));
 		notifies->interp = interp;
 		Tcl_InitHashTable(&notifies->notify_hash, TCL_STRING_KEYS);
+		notifies->conn_loss_cmd = NULL;
 		notifies->next = connid->notify_list;
 		connid->notify_list = notifies;
 		Tcl_CallWhenDeleted(interp, PgNotifyInterpDelete,
@@ -1969,3 +1973,84 @@ Pg_listen(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
 	ckfree(caserelname);
 	return TCL_OK;
 }
+
+/***********************************
+Pg_on_connection_loss
+	create or remove a callback request for unexpected connection loss
+
+ syntax:
+   pg_on_connection_loss conn ?callbackcommand?
+
+   With a third arg, creates or changes the callback command for
+   connection loss; without, cancels the callback request.
+
+   Callbacks can occur whenever Tcl is executing its event loop.
+   This is the normal idle loop in Tk; in plain tclsh applications,
+   vwait or update can be used to enter the Tcl event loop.
+***********************************/
+int
+Pg_on_connection_loss(ClientData cData, Tcl_Interp *interp, int argc, char *argv[])
+{
+	char	   *callback = NULL;
+	Pg_TclNotifies *notifies;
+	Pg_ConnectionId *connid;
+	PGconn	   *conn;
+
+	if (argc < 2 || argc > 3)
+	{
+		Tcl_AppendResult(interp, "wrong # args, should be \"",
+						 argv[0], " connection ?callback?\"", 0);
+		return TCL_ERROR;
+	}
+
+	/*
+	 * Get the command arguments.
+	 */
+	conn = PgGetConnectionId(interp, argv[1], &connid);
+	if (conn == (PGconn *) NULL)
+		return TCL_ERROR;
+
+	if ((argc > 2) && *argv[2])
+	{
+		callback = (char *) ckalloc((unsigned) (strlen(argv[2]) + 1));
+		strcpy(callback, argv[2]);
+	}
+
+	/* Find or make a Pg_TclNotifies struct for this interp and connection */
+
+	for (notifies = connid->notify_list; notifies; notifies = notifies->next)
+	{
+		if (notifies->interp == interp)
+			break;
+	}
+	if (notifies == NULL)
+	{
+		notifies = (Pg_TclNotifies *) ckalloc(sizeof(Pg_TclNotifies));
+		notifies->interp = interp;
+		Tcl_InitHashTable(&notifies->notify_hash, TCL_STRING_KEYS);
+		notifies->conn_loss_cmd = NULL;
+		notifies->next = connid->notify_list;
+		connid->notify_list = notifies;
+		Tcl_CallWhenDeleted(interp, PgNotifyInterpDelete,
+							(ClientData) notifies);
+	}
+
+	/* Store new callback setting */
+
+	if (notifies->conn_loss_cmd)
+		ckfree((void *) notifies->conn_loss_cmd);
+	notifies->conn_loss_cmd = callback;
+
+	if (callback)
+	{
+		/*
+		 * Start the notify event source if it isn't already running. The
+		 * notify source will cause Tcl to watch read-ready on the
+		 * connection socket, so that we find out quickly if the
+		 * connection drops.
+		 */
+		PgStartNotifyEventSource(connid);
+	}
+
+	return TCL_OK;
+}
